"""skillmat specific entities

:organization: Logilab
:copyright: 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from cubicweb import _
from cubicweb.predicates import is_instance
from cubicweb.entities import AnyEntity, adapters

from cubicweb_folder.entities import Folder


class Masters(AnyEntity):
    __regid__ = 'Masters'
    fetch_attrs = ('rate', 'foruser', 'skill')

    def dc_title(self):
        return u'%s: %s' % (self.skill[0].name, self.rate)

    def dc_long_title(self):
        return u'%s / %s' % (self.foruser[0].dc_title(), self.dc_title())


class Technology(AnyEntity):
    __regid__ = 'Technology'
    fetch_attrs = ('name',)

    def skill_categories(self):
        # XXX what if we have a skill tree with depth > 1
        return self.filed_under


class Skill(Folder):  # XXX

    def skill_categories(self):
        return [self]


class Talk(AnyEntity):
    __regid__ = 'Talk'
    fetch_attrs = ('subject', 'description', 'talktime')

    def dc_title(self):
        return u'%s' % (self.subject)


SKILLS = {0: _("I don't want to learn this technology"),
          1: _('I know nothing about this technology'),
          2: _('Just beginning whith this technology'),
          3: _('I start on this technology, and I can help'),
          4: _('I am doing well, selfworking'),
          5: _('I can conduct a training course')}


class TechnologyITreeAdapter(adapters.ITreeAdapter):
    __select__ = is_instance('Technology')
    tree_relation = 'filed_under'

    def children(self, entities=True, sametype=False):
        if entities:
            return ()
        return self._cw.empty_rset()



"""template-specific forms/views/actions/components"""
import six
from cubicweb_web.view import EntityAdapter
from cubicweb.predicates import is_instance
from cubicweb_web.views import uicfg

from cubicweb_skillmat.entities import SKILLS

RATEVOCAB = [(l, six.text_type(v)) for v, l in sorted(SKILLS.items())]
uicfg.autoform_field_kwargs.tag_attribute(
    ('Masters', 'rate'),
    {'choices': RATEVOCAB,
     'internationalizable': True})

uicfg.autoform_section.tag_subject_of(('Talk', 'attended_by', '*'), 'main', 'attributes')
uicfg.autoform_section.tag_subject_of(('Talk', 'attended_by', '*'), 'muledit', 'attributes')


class TalkICalendarable(EntityAdapter):
    __regid__ = 'ICalendarable'
    __select__ = is_instance('Talk')

    @property
    def start(self):
        return self.entity.talktime

    @property
    def stop(self):
        return self.entity.talktime



from yams.buildobjs import (EntityType, SubjectRelation, RelationDefinition,
                            String, Datetime, Int)
from yams.constraints import IntervalBoundConstraint

from cubicweb.schema import ERQLExpression

from cubicweb_folder.schema import Folder

from cubicweb import _


class Masters(EntityType):  # XXX find a suitable name
    """relation promoted to EntityType to store the `rate` attribute
    """
    __permissions__ = {'read': ('managers', 'users'),
                       'add': ('managers', ERQLExpression('X foruser U')),
                       'update': ('managers', 'owners'),
                       'delete': ('managers',),
                       }
    rate = Int(required=True, constraints=[IntervalBoundConstraint(0, 5)])
    foruser = SubjectRelation('CWUser', cardinality='1*')
    skill = SubjectRelation(('Technology', 'Folder'), cardinality='1*')


class Technology(Folder):
    pass


class Talk(EntityType):
    """relation promoted to EntityType to store the `subject' attribute
    """
    subject = String(required=True, fulltextindexed=True, indexed=True, maxsize=32)
    description = String(fulltextindexed=True,
                         description=_('more detailed subject description'))
    talktime = Datetime(description=_('estimated presentation date'))
    presented_by = SubjectRelation('CWUser', cardinality='?*')
    attended_by = SubjectRelation('CWUser', cardinality='**')
    talks_about = SubjectRelation(('Technology', 'Folder'), cardinality='+*')


class comments(RelationDefinition):
    subject = 'Comment'
    object = 'Talk'
    cardinality = '1*'
    composite = 'object'

# extend CWUser


class wishes_to_learn(RelationDefinition):
    subject = 'CWUser'
    object = ('Technology', 'Folder')


class interested_in(RelationDefinition):
    subject = 'CWUser'
    object = 'Talk'



import unittest

from cubicweb.devtools.testlib import CubicWebTC


class HookTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            self.skill = cnx.create_entity('Folder', name=u'python')
            self.u = self.create_user(cnx, u'joe', password=u'joe')
            cnx.commit()

    def test_owned_by_auto_set(self):
        """test owned_by insertion when all is inserted in 1 query"""
        with self.admin_access.cnx() as cnx:
            skilleid = self.skill.eid
            meid = cnx.execute('INSERT Masters M: M rate 3, M foruser U, M skill S WHERE '
                               'U eid %(u)s, S eid %(s)s', {'u': self.u.eid, 's': skilleid})[0][0]
            owners = cnx.execute('Any L WHERE U login L, M owned_by U, M eid %s' % meid)
            cnx.commit()
            self.assertSetEqual(set(row[0] for row in owners), set(('admin', 'joe')))

    def test_owned_by_auto_set_2(self):
        """test owned_by insertion when all is inserted in separate queries"""
        with self.admin_access.cnx() as cnx:
            skilleid = self.skill.eid
            meid = cnx.execute('INSERT Masters M: M rate 3, M foruser U, M skill S WHERE '
                               'S eid %(s)s', {'s': skilleid})[0][0]
            # should only owned by admin at this point
            owners = cnx.execute('Any L WHERE U login L, M owned_by U, M eid %s' % meid)
            cnx.commit()
            self.assertSetEqual(set(row[0] for row in owners), set(('admin',)))
            cnx.execute('SET M foruser U WHERE M eid %s, U eid %s' % (meid, self.u.eid))
            # should now be owned by admin and targeted user
            owners = cnx.execute('Any L WHERE U login L, M owned_by U, M eid %s' % meid)
            cnx.commit()
            self.assertSetEqual(set(row[0] for row in owners), set(('admin', 'joe')))

    def test_owned_by_auto_set_3(self):
        """test owned_by insertion when targeted user is the creator
        (all in one query)
        """
        with self.new_access(u'joe').cnx() as cnx:
            skilleid = self.skill.eid
            meid = cnx.execute('INSERT Masters M: M rate 3, M foruser U, M skill S WHERE '
                               'U eid %(u)s, S eid %(s)s', {'u': self.u.eid, 's': skilleid})[0][0]
            owners = cnx.execute('Any L WHERE U login L, M owned_by U, M eid %s' % meid)
            cnx.commit()
            self.assertSetEqual(set(row[0] for row in owners), set((u'joe',)))

    def test_owned_by_auto_set_4(self):
        """test owned_by insertion when targeted user is the creator"""
        with self.new_access(u'joe').cnx() as cnx:
            skilleid = self.skill.eid
            meid = cnx.execute('INSERT Masters M: M rate 3, M skill S WHERE '
                               'S eid %(s)s', {'s': skilleid})[0][0]
            # should only owned by admin at this point
            owners = cnx.execute('Any L WHERE U login L, M owned_by U, M eid %s' % meid)
            self.assertSetEqual(set(row[0] for row in owners), set(('joe',)))
            cnx.execute('SET M foruser U WHERE M eid %s, U eid %s' % (meid, self.u.eid))
            # should now be owned by admin and targeted user
            owners = cnx.execute('Any L WHERE U login L, M owned_by U, M eid %s' % meid)
            self.assertSetEqual(set(row[0] for row in owners), set(('joe',)))


if __name__ == '__main__':
    unittest.main()



# pylint: disable-msg=W0622
"""cubicweb-skillmat application packaging information"""

modname = 'skillmat'
distname = 'cubicweb-skillmat'

numversion = (0, 12, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
description = 'skill matrix component for the CubicWeb framework'
author = 'Logilab'
author_email = 'contact@logilab.fr'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {'cubicweb': ">=3.38.0,<3.39.0",
               'cubicweb-folder': ">=2.1.0,<2.2.0",
               'cubicweb-comment': ">=2.0.0,<2.1.0",
               "pandas": ">=1.5.0,<1.6.0"}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
    ]



"""skillmat primary views

:organization: Logilab
:copyright: 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from logilab.mtconverter import xml_escape

from cubicweb.predicates import is_instance
from cubicweb_web.views import primary

from cubicweb_folder.views import FolderPrimaryView


class SkillPrimaryView(FolderPrimaryView):

    def cell_call(self, row, col):
        super(SkillPrimaryView, self).cell_call(row, col)
        skill = self.cw_rset.get_entity(row, col)
        rql = ('Any U,L,R ORDERBY L WITH U,L,R BEING ( '
               # get username and skill for those who specified it
               '   (Any U,L,R WHERE U login L, M foruser U, M rate R, '
               '    M skill S, S eid %(s)s) '
               ' UNION '  # union users for who the skill is unknown
               '   (Any U,L,-1 WHERE U login L, U is CWUser, '
               '    NOT EXISTS(M skill S, S eid %(s)s, M foruser U))  )')
        rset = self._cw.execute(rql, {'s': skill.eid})
        if not rset:
            self.warning('no skill information found for %s', skill.dc_title())
            return
        self._cw.add_css('cubes.skillmat.css')
        self.w(u'<table class="matrix">')
        self.w(u'<tr>')
        for row in range(len(rset)):
            self.w(u'<th>%s</th>' % self._cw.view('oneline', rset, row=row, col=0))
        self.w(u'</tr>\n<tr>')
        width = 100 / len(rset)
        for ueid, _, rate in rset:
            if rate == -1:
                linkto = ('foruser:%s:subject' % ueid,
                          'skill:%s:subject' % skill.eid)
                url = xml_escape(self._cw.build_url('add/Masters', __linkto=linkto))
                self.w(u'<td style="width: %s%%"><a href="%s">%s</a></td>' % (
                    width, url, self._cw._('???')))
            else:
                linkto = ('foruser:%s:subject' % ueid,
                          'skill:%s:subject' % skill.eid)
                url = xml_escape(self._cw.build_url('change/Masters', __linkto=linkto))
                self.w(u'<td style="width: %s%%"><a href="%s">%s</a></td>' % (
                    width, url, self._cw._('???')))

                self.w(u'<td style="width: %s%%" class="l%s">&nbsp;</td>' % (width, rate))
        self.w(u'</tr>')
        self.w(u'</table>')


class TalkPrimaryView(primary.PrimaryView):
    __select__ = is_instance('Talk',)

    def render_entity_title(self, entity):
        self.w(u'<h1>%s</h1>' % xml_escape(entity.dc_title()))

    def render_entity_relations(self, entity):
        if entity.related('interested_in', 'object'):
            self.w(self._cw._('interested people :'))
            self.wview('list', entity.related('interested_in', 'object'), 'null')
        if entity.related('presented_by'):
            self.w(self._cw._('presented by :'))
            self.wview('list', entity.related('presented_by'), 'null')


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__, (SkillPrimaryView,))
    vreg.register_and_replace(SkillPrimaryView, FolderPrimaryView)



import unittest

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb import Unauthorized


class SecurityTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.cnx() as cnx:
            self.skill = cnx.create_entity("Folder", name=u'python')
            self.u1 = self.create_user(cnx, u'joe', password=u'joe')
            self.u2 = self.create_user(cnx, u'jack', password=u'jack')
            self.master = cnx.create_entity('Masters', rate=3, foruser=self.u1,
                                            skill=self.skill)
            cnx.commit()

    def test_admin_can_update(self):
        with self.admin_access.cnx() as cnx:
            eid = self.master.eid
            cnx.execute('SET M rate 4 WHERE M eid %s' % eid)
            cnx.commit()
            self.assertEqual(cnx.execute('Any R WHERE M eid %s, M rate R' % eid)[0][0], 4)

    def test_owner_can_update(self):
        with self.new_access(u'joe').cnx() as cnx:
            eid = self.master.eid
            cnx.execute('SET M rate 4 WHERE M eid %s' % eid)
            cnx.commit()
            self.assertEqual(cnx.execute('Any R WHERE M eid %s, M rate R' % eid)[0][0], 4)

    def test_others_cannot_update(self):
        with self.new_access(u'jack').cnx() as cnx:
            eid = self.master.eid
            cnx.execute('SET M rate 4 WHERE M eid %s' % eid)
            self.assertRaises(Unauthorized, cnx.commit)
            self.assertEqual(cnx.execute('Any R WHERE M eid %s, M rate R' % eid)[0][0], 3)


if __name__ == '__main__':
    unittest.main()



"""skillmat boxes

:organization: Logilab
:copyright: 2003-2010 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"

from cubicweb.predicates import is_instance, match_user_groups
from cubicweb_web import component


class EditWishesBox(component.EditRelationCtxComponent):
    __regid__ = 'edit-wishes'
    __select__ = (component.EditRelationCtxComponent.__select__
                  & is_instance('CWUser') & match_user_groups('owners'))

    rtype = 'wishes_to_learn'
    target = 'object'

    def mastered_skills(self, user, threshold=5):
        """return the list of mastered skills
        """
        rql = 'Any S,N WHERE M skill S, S name N, M rate >= %(r)s, M foruser U, U eid %(u)s'
        rset = self._cw.execute(rql, {'r': threshold, 'u': user.eid})
        return rset.entities()

    def unrelated_entities(self, euser):
        """filter skills for which the user already knows"""
        entities = super(EditWishesBox, self).unrelated_entities(euser)
        mastered = set(entity.eid for entity in self.mastered_skills(euser))
        return [entity for entity in entities if entity.eid not in mastered]


class EditAttendedByBox(component.EditRelationCtxComponent):
    __regid__ = 'edit-attended-by'
    __select__ = component.EditRelationCtxComponent.__select__ & is_instance('Talk')

    rtype = 'attended_by'
    target = 'object'

    # Todo : replace + et - in this box by a comprehensive label

    # def w_related(self, box, entity):
    #     """appends existing relations to the `box`"""
    #     rql = 'DELETE S %s O WHERE S eid %%(s)s, O eid %%(o)s' % self.rtype
    #     related = self.related_entities(entity)
    #     for etarget in related:
    #         box.append(self.box_item(entity, etarget, rql, u'assiste'))
    #     return len(related)

    # def w_unrelated(self, box, entity):
    #     """appends unrelated entities to the `box`"""
    #     rql = 'SET S %s O WHERE S eid %%(s)s, O eid %%(o)s' % self.rtype
    #     for etarget in self.unrelated_entities(entity):
    #         box.append(self.box_item(entity, etarget, rql, u'n\'assite pas'))



add_package('ecomment')
# Talk entity type is added automatically by the add_package command
# but attribute order doesn't seem to be right, use synchronize to
# reset the order
print "please, restart the shell and execute synchronize_eschema('Talk')"
# synchronize_eschema('Talk')



"""this contains the server-side objects"""

from cubicweb.server.hook import Hook, match_rtype


class OnNewMastersHook(Hook):
    """adds owned_by relation betweem the Masters and the targeted user
    """
    __regid__ = 'skillmat_set_owner_when_set_foruser'
    __select__ = Hook.__select__ & match_rtype('foruser')
    events = ('after_add_relation',)

    def __call__(self):
        self._cw.execute('SET M owned_by U WHERE M eid %(m)s, U eid %(u)s, NOT M owned_by U',
                         {'m': self.eidfrom, 'u': self.eidto})



import unittest
from cubicweb.devtools.testlib import AutomaticWebTest


class AutomaticWebTest(AutomaticWebTest):
    def to_test_etypes(self):
        return set(('Masters', 'Technology', 'Talk', 'Folder', 'CWUser'))

    def list_startup_views(self):
        return ('index',)


if __name__ == '__main__':
    unittest.main()



from logilab.common.decorators import cached
from logilab.mtconverter import xml_escape

import pandas as pd
from cubicweb import _
from cubicweb_web.view import EntityView
from cubicweb.predicates import is_instance, match_view
from cubicweb.uilib import toggle_action
from cubicweb_web import action
from cubicweb_web.views.csvexport import CSVMixIn

from cubicweb_skillmat.entities import SKILLS


class SkillMatrixView(EntityView):
    __regid__ = 'matrix'
    __select__ = is_instance('CWUser')

    def call(self):
        skills = self.skills
        if not skills:
            self.w(self._cw._('no skill definition found'))
            return
        # prefetch every skill for users in the resultset
        self.masters_dict = self.prefetch_skills()
        self._cw.add_css('cubes.skillmat.css')
        divid = 'appMsg%s' % id(self)
        # FIXME: make the applmessages component flexible enough to use it
        msg = self._cw._('you can update your skill information by clicking on the corresponding cell')
        self.w(u'''<div class="appMsg" id="%s" onclick="%s">
        <div class="message">%s</div>
        </div>
        ''' % (divid, toggle_action(divid), xml_escape(msg)))
        self.w(u'<table class="matrix">')
        self.w(u'<tr>')
        self.w(u'<th style="width: 10%;">&nbsp;</th>')  # users column
        width = int(90 / len(skills))
        for skill in skills:
            self.w(u'<th style="width: %s%%;">%s</th>' %
                   (width, skill.view('oneline')))
        self.w(u'</tr>')
        for row in range(len(self.cw_rset)):
            self.cell_call(row, 0)
        self.w(u'</table>')
        self._build_help()

    def _build_help(self):
        """builds an HTML section explaining what each colour means"""
        self.w(u'<br />')
        self.w(u'<table class="matrixHelp">')
        celltemplatelabel = u'<td class="l%s rateInfo">&nbsp;</td>'
        matrix_label = dict(SKILLS)
        matrix_label[-1] = self._cw._(u'no information for this technology')
        for rate, label in sorted(matrix_label.items()):
            self.w(u'<tr>')
            self.w(celltemplatelabel % rate)
            self.w(u'<td>%s</td>' % self._cw._(label))
            self.w(u'</tr>')
        self.w(u'</table>')
        self.w(u'<br />')

    def cell_call(self, row, col):
        user = self.cw_rset.get_entity(row, col)
        ruser = self._cw.user
        relpath = self._cw.relative_path(includeparams=False)
        self.w(u'<tr>')
        self.w(u'<th style="width: 10%%;">%s</th>' % user.view('oneline'))
        skills = self.skills
        width = int(90 / len(skills))
        celltemplate1 = u'<td %%sstyle="width: %s%%%%">[<a href="%%s">+</a>]</td>' % width
        celltemplate2 = u'<td %%sstyle="width: %s%%%%">&nbsp;</td>' % width
        for skill in skills:
            user_can_update = (user.eid == ruser.eid) or ruser.is_in_group('managers')
            try:
                master = self.masters_dict[user.eid][skill.name]
            except KeyError:  # skill or even user might not be present
                master = None
            if master is None:
                if user_can_update:
                    url = self.add_masters_url(user, skill.eid, __redirectpath=relpath)
                    self.w(celltemplate1 % (u'', xml_escape(url)))
                else:
                    self.w(celltemplate2 % u'')
            else:
                attrs = u'class="l%s" ' % master.rate
                if user_can_update:
                    url = master.absolute_url(vid='edition', __redirectpath=relpath)
                    self.w(celltemplate1 % (attrs, xml_escape(url)))
                else:
                    self.w(celltemplate2 % attrs)
        self.w(u'</tr>')

    def add_masters_url(self, user, skilleid, **kwargs):
        linkto = ('foruser:%s:subject' % user.eid,
                  'skill:%s:subject' % skilleid)
        return self._cw.build_url('add/Masters', __linkto=linkto, **kwargs)

    @property
    @cached
    def skills(self):
        rset = self._cw.execute('Any F,N ORDERBY N WHERE F is Folder, F name N')
        return list(rset.entities())

    def prefetch_skills(self):
        """computes once and for all each skill rate for each user.

        :returns: a dict mapping user eid to a subdict mapping skillname
                  to masters object
        """
        rset = self._cw.execute('Any U,M,R,N WHERE M is Masters, M rate R, M foruser U, '
                                'M skill S, S is Folder, S name N, U eid IN (%s)'
                                % ','.join(str(row[0]) for row in self.cw_rset))
        masters_dict = {}
        for rowidx, (ueid, meid, rate, skillname) in enumerate(rset):
            masters = rset.get_entity(rowidx, 1)
            masters_dict.setdefault(ueid, {})[skillname] = masters
        return masters_dict


class SkillMatrixExport(CSVMixIn, EntityView):
    __regid__ = 'skillmat-export'
    __select__ = is_instance('CWUser')

    def call(self):
        users_skills = self._cw.execute(
            "Any L,R,N WHERE M is Masters, M rate R, M foruser U, "
            "M skill S, S is Folder, S name N, U login L, U eid IN (%s)"
            % ','.join(str(row[0]) for row in self.cw_rset)
        )
        dataframe = pd.DataFrame(
            users_skills,
            columns=["login", "rate", "skill_name"]
        )
        self.w(
            dataframe.pivot_table(
                index="login", columns="skill_name", values="rate",
            ).to_csv()
        )


class ExportMatrixAction(action.Action):
    __regid__ = 'skillmat-export-action'
    __select__ = (
        action.Action.__select__
        & is_instance("CWUser")
        & match_view("matrix")
    )

    category = "mainactions"
    title = _("export SkillMatrix")

    def url(self):
        return self._cw.build_url(rql=self.cw_rset.rql, vid="skillmat-export")
